/*
** Module   : ArrayAssigner.java
** Abstract : automates the assignment of each element of a BDT array
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description----------------------------
** 001 GES 20051104   @23229 Provides a simple helper to automate the
**                           assignment of each element of a BaseDataType
**                           array based on an array input.
** 002 GES 20060410   @25444 All elements can be set to the same scalar
**                           value (new signature).
** 003 GES 20060412   @25486 Added a setUnknown() method.
** 004 GES 20060425   @25707 Added a new assign() signature.
** 005 ECF 20060428   @25851 Added safety code. If target array element is
**                           null, duplicate corresponding source to
**                           initialize it.
** 006 GES 20080911   @39793 Added signatures to allow primitives to be
**                           assigned from.
** 007 GES 20090625   @42956 Fix and helpers to avoid ArrayIndexOutOfBounds.
**                           This implements a new delegation model to hide
**                           all array variable dereferencing (for assignment
**                           or simple access). The helpers can thus translate
**                           or avoid Java level AIOOB exceptions and behave
**                           the way that the 4GL requires.
** 008 CA  20130209          Added support for the EXTENT statement and output EXTENT parameters.
** 009 CA  20130315          Added missing assignSingle and assignMulti APIs, for to the cases
**                           when the rvalue is a resource, as in some cases the :HANDLE attribute
**                           conversion emits the WrappedResource instance instead of handle.
** 010 OM  20130418          Switched to long/int64 to support 64-bit wide indexes and values. 
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.security.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Provides a simple helper to automate the assignment of each element of a 
 * {@link BaseDataType} array based on 2 array inputs.
 */
public class ArrayAssigner
{
   /** Context-local data container. */
   private static ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         TransactionManager.registerScopeable(wa);
         return wa;
      };
   };

   /**
    * Ensure all the context-local data has initialized (to be able to listen for scope 
    * notifications).
    */
   static void load()
   {
      // needed to ensure the variable is loaded, so it listens for scope notifications
      local.get();
   }

   /**
    * Assigns (mutates the content of) each matching element of a target
    * array based on the corresponding element of the source array.  The
    * source and target arrays must be assignable to each other (must be
    * assignment-compatible). The copy operation will stop as soon as the end
    * of either array is encountered. If either array is longer than the other,
    * then the copy will only occur for the overlapping portion. The copy
    * always starts at element 0 and finishes at the lesser of source length or
    * target length.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The array from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, BaseDataType[] source) 
   throws ErrorConditionException,
          NullPointerException
   {
      boolean dynamicTarget = isDynamicArray(target);
      boolean dynamicSource = isDynamicArray(source);

      if (dynamicTarget     && 
          target.length > 0 && 
          (source.length == 0 ||
           source.length != target.length))
      {
         final String err = "Indeterminate extent is already fixed to a dimension of %d";
         String errMsg = String.format(err, target.length);
         ErrorManager.recordOrThrowError(13738, errMsg, false);
         
         return;
      }
      if (!dynamicTarget)
      {
         if (source.length == 0)
         {
            final String errMsg = "Uninitialized array used as source of assignment";
            ErrorManager.recordOrThrowError(14906, errMsg, false);
            
            return;
         }
         else if (source.length != target.length)
         {
            final String errMsg = 
               "Whole-array assignment target and source must have the same extent unless " +
               "the target is indeterminate";
            ErrorManager.recordOrThrowError(14905, errMsg, false);
            
            return;
         }
      }
      
      if (target.length == 0 && source.length > 0)
      {
         target = resize(target, source.length);
      }

      for (int i = 0; (i < target.length && i < source.length); i++)
      {
         if (target[i] == null)
         {
            target[i] = source[i].duplicate();
         }
         else
         {
            (target[i]).assign(source[i]);
         }
      }
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new int64(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, double source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new decimal(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, String source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new character(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, boolean source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new logical(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, BaseDataType source) 
   throws ErrorConditionException,
          NullPointerException
   {
      if (target.length == 0 && TransactionManager.blockInitialized())
      {
         // TODO: for some reason, this error message doesn't end in a dot in 4GL
         String msg = "Invalid assignment to an unfixed indeterminate extent";
         ErrorManager.recordOrThrowError(11389, msg, false);
         
         return;
      }
      
      for (int i = 0; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = source.duplicate();
         }
         else
         {
            (target[i]).assign(source);
         }
      }
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, WrappedResource source) 
   throws ErrorConditionException,
          NullPointerException
   {
      if (target.length == 0 && TransactionManager.blockInitialized())
      {
         // TODO: for some reason, this error message doesn't end in a dot in 4GL
         String msg = "Invalid assignment to an unfixed indeterminate extent";
         ErrorManager.recordOrThrowError(11389, msg, false);
         
         return;
      }
      
      for (int i = 0; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = new handle(source);
         }
         else
         {
            (target[i]).assign(new handle(source));
         }
      }
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new int64(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, double source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new decimal(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, String source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new character(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, boolean source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new logical(source));
   }
   
   /**           
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, BaseDataType source)
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = idx; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = source.duplicate();
         }
         else
         {
            (target[i]).assign(source);
         }
      }
   }   
   
   /**           
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[]  target, int idx, WrappedResource source) 
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = idx; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = new handle(source);
         }
         else
         {
            (target[i]).assign(new handle(source));
         }
      }
   }   
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new int64(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, double source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new decimal(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, boolean source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new logical(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, String source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new character(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, BaseDataType source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), source);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, WrappedResource source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new handle(source));
   }
      
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new int64(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, double source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new decimal(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, boolean source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new logical(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, String source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new character(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, BaseDataType source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleWorker(target, index, source);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, WrappedResource source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleWorker(target, index, new handle(source));
   }
   
   /**
    * Dereference the given array using the given 1-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    */
   public static <B extends BaseDataType> B subscript(B[] array, long index)
   {
      int idx = subscriptHelper(array, index);
      return subscriptWorker(array, idx);
   }
   
   /**
    * Dereference the given array using the given 1-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    */
   public static <B extends BaseDataType> B subscript(B[] array, NumberType index)
   {
      int idx = subscriptHelper(array, index, false);
      return subscriptWorker(array, idx);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the unknown value.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingleUnknown(BaseDataType[] target, long index)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleUnknown(target, new int64(index));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the unknown value.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingleUnknown(BaseDataType[] target, NumberType index)
   throws ErrorConditionException,
          NullPointerException
   {
      Class<? extends BaseDataType> cls = 
         (Class<? extends BaseDataType>) target.getClass().getComponentType();

      assignSingleWorker(target, index, BaseDataType.generateUnknown(cls));
   }
   
   /**                     
    * Sets all elements of a target array to the <code>unknown value</code>.
    *
    * @param    target
    *           The array to be modified. Must not be <code>null</code>.
    */
   public static void assignUnknown(BaseDataType[] target)
   {
      if (target.length == 0 && TransactionManager.blockInitialized())
      {
         // TODO: for some reason, this error message doesn't end in a dot in 4GL
         String msg = "Invalid assignment to an unfixed indeterminate extent";
         ErrorManager.recordOrThrowError(11389, msg, false);
         
         return;
      }
      
      for (int i = 0; i < target.length; i++)
      {
         target[i].setUnknown();
      }
   }
   
   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero, then an error is raised.  Else, all the elements
    * of the array will be set to unknown.
    * 
    * @param    array
    *           The dynamic-extent array to be resized.
    * @param    newSize
    *           The array's new size.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, int newSize)
   {
      return resize(array, new integer(newSize));
   }

   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero (and the newSize is not unknown), then an error is 
    * raised.  Else, the array will be resized and all the elements will be set to the default
    * value.
    * <p>
    * The new extent value must be from the [1..28000] interval.
    * 
    * @param    array
    *           The dynamic-extent array to be resized. Must not be null. If the array was not
    *           resized yet, must be 0-length.
    * @param    newSize
    *           The array's new size.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, int64 newSize)
   {
      long size = newSize.isUnknown() ? 0 : newSize.longValue();
      if (!newSize.isUnknown())
      {
         if (array.length != 0)
         {
            final String msg = "Indeterminate extent is already fixed to a dimension of %d";
            ErrorManager.recordOrThrowError(13738, String.format(msg, array.length), false);
            return array;
         }
   
         if (size <= 0 || size > 28000)
         {
            //size < 1:  Invalid value specified for the EXTENT statement. (13737)
            //size > 28000: Array extent may not be greater than 28000 (3350)
            final String msg = "Indeterminate extent dimension %d is out of range 1 to 28000";
            ErrorManager.recordOrThrowError(11385, String.format(msg, size), false);
            return array;
         }
      }
      
      Class<? extends BaseDataType> cls = 
         (Class<? extends BaseDataType>) array.getClass().getComponentType();
      T[] copy = (T[]) Array.newInstance(cls, (int) size);
      for (int i = 0; i < copy.length; i++)
      {
         copy[i] = (T) BaseDataType.generateDefault(cls);
      }
      
      WorkArea wa = local.get();
      // deregister it
      wa.deregister(array);
      // register new reference
      wa.register(copy);

      return copy;
   }

   /**
    * Get a copy of the given array.
    * 
    * @param    array
    *           The array for which a copy is needed.
    * 
    * @return   The copy.
    */
   public static <T extends BaseDataType> T[] copyOf(T[] array)
   {
      Class<? extends BaseDataType> cls = 
               (Class<? extends BaseDataType>) array.getClass().getComponentType();
   
      T[] copy = (T[]) Array.newInstance(cls, array.length);
      for (int i = 0; i < copy.length; i++)
      {
         copy[i] = (T) BaseDataType.generateUnknown(cls);
         copy[i].assign(array[i]);
      }
      
      return copy;
   }

   /**
    * Get the length of the given dynamic-extent variable.
    * 
    * @param    array
    *           The dynamic-extent variable.
    *           
    * @return   If the array's length is zero, return unknown. Else, the actual size.
    */
   public static integer lengthOf(BaseDataType[] array)
   {
      return array.length == 0 ? new integer() : new integer(array.length);
   }

   /**
    * Register the given array as a dynamic-array.
    * 
    * @param    array
    *           The dynamic-array to register.
    */
   public static void registerDynamicArray(BaseDataType[] array)
   {
      local.get().register(array);
   }
   
   /**
    * This method returns true for any real dynamic-extent array (regardless if it had its
    * size set).
    * 
    * @param    array
    *           The array to be checked.
    *           
    * @return   See above.
    */
   static boolean isDynamicArray(BaseDataType[] array)
   {
      return local.get().allDynamicArrays.contains(array);
   }

   /**
    * Dereference the given array using the given 0-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    idx
    *           The 0-based index into the array. No bounds checking is done.
    */
   private static <B extends BaseDataType> B subscriptWorker(B[] array, int idx)
   {
      if (idx == -2)
      {
         // this case means the index is unknown, so we need to return unknown
         return (B) BaseDataType.generateUnknown(array.getClass().getComponentType());
      }
      
      if (idx == -1 || ErrorManager.isPendingError())
      {
         Class<?> cls = array.getClass().getComponentType();
         
         // silently return unknown value (safest thing)
         return (B) BaseDataType.generateUnknown(cls);
      }
      
      return array[idx];
   }
   
   /**
    * Generate a valid 0-based integral subscript from the 1-based numeric 
    * value given. Bounds checking is done based on the target array length.
    *
    * @param    array
    *           The array to dereference.
    * @param    _idx
    *           The 1-based index to be checked and modified.
    * @param    assign
    *           <code>true</code> if the extent var in on the left-side of the assign.
    *
    * @return   The 0-based integral subscript, -2 if the index is the
    *           <code>unknown value</code> or -1 if the index is out of bounds
    *           in silent error mode.
    *
    * @throws   ErrorConditionException
    *           If the index is out of bounds.
    */
   private static int subscriptHelper(BaseDataType[] array, NumberType _idx, boolean assign)
   throws ErrorConditionException
   { 
      if (_idx.isUnknown())
      {
         // no error is generated if the index is unknown, but unknown needs to be returned
         return -2;
      }

      // 4GL lets the the 33 MSB overflow, computing the array index using last 31 bits
      int idx = (int)(_idx.longValue() & 0x7FFFFFFFL);
      if (idx < 0)
      {
         String spec   = "Array subscript %d is out of range";
         String errmsg = String.format(spec, idx);
         
         ErrorManager.recordOrThrowError(26, errmsg);
         return idx - 1;
      }

      int len = array.length;
      if (len == 0)
      {
         // dynamic-extent, not initialized
         String msg = "Invalid runtime use of indeterminate extent.  " +
                      "It must be set to a fixed dimension";
         
         // TODO: if we are in message mode, it seems like the error is shown, but not thrown
         ErrorManager.recordOrThrowError(11387, msg, false);

         // we need to return -1
         return -1;
      }
      else if (idx <= 0 || idx - 1 >= len)
      {
         // error out!
         if (isDynamicArray(array) && (!assign || idx - 1 >= len))
         {
            String spec = 
               "Array subscript %d is out of range.  The indeterminate extent is fixed to a " +
               "dimension of %d";
            String errmsg = String.format(spec, idx, len);
            
            ErrorManager.recordOrThrowError(11388, errmsg, false);
         }
         else
         {
            String spec   = "Array subscript %d is out of range";
            String errmsg = String.format(spec, idx);
            
            ErrorManager.recordOrThrowError(26, errmsg);
         }

         // we need to return -1
         return -1;
      }
      
      return idx - 1;
   }
   
   /**
    * Generate a valid 0-based integral subscript from the 1-based numeric 
    * value given. Bounds checking is done based on the target array length.
    *
    * @param    array
    *           The array to dereference.
    * @param    idx
    *           The 1-based index to be checked and modified.
    *
    * @return   The 0-based integral subscript or -1 if the index is out
    *           of bounds in silent error mode.
    *
    * @throws   ErrorConditionException
    *           If the index is out of bounds.
    */
   private static int subscriptHelper(BaseDataType[] array, long idx)
   throws ErrorConditionException
   {
      return subscriptHelper(array, new int64(idx), false);
   }

   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   private static void assignSingleWorker(BaseDataType[] target,
                                          NumberType     index,
                                          BaseDataType   source)
   {

      boolean unknown = index.isUnknown();
      boolean dynamic = isDynamicArray(target);
      if (!unknown && target.length == 0)
      {
         // dynamic-extent, not initialized, and tried to assign an element
         if (index.intValue() <= 0)
         {
            String spec   = "Array subscript %d is out of range";
            String errmsg = String.format(spec, index.intValue());
            
            ErrorManager.recordOrThrowError(26, errmsg);
         }
         else
         {
            // TODO: for some reason, this error message doesn't end in a dot in 4GL
            String msg = "Invalid assignment to an unfixed indeterminate extent";
            ErrorManager.recordOrThrowError(11389, msg, false);
         }
         return;
      }
      
      int idx = -1;
      
      try
      {
         idx = subscriptHelper(target, index, true);
         if (idx == -2)
         {
            // in case the index is unknown, show only this message
            ErrorManager.recordOrThrowError(26, "Array subscript 0 is out of range");
            return;
         }
      }
      
      catch (ErrorConditionException ece)
      {
         // re-throw the error if in dynamic mode or if the index is less than zero
         if (dynamic || (!index.isUnknown() && index.intValue() < 0))
            throw ece;
         
         // eat the first error, we will rethrow shortly
      }
      
      if (idx == -1 || ErrorManager.isPendingError())
      {
         if (!dynamic)
         {
            // only non-dynamic arrays show this error (double space before Field is indented)
            ErrorManager.recordOrThrowError(142, "Unable to update  Field");
         }
         
         // silently return
         return;
      }
      
      if (target[idx] == null)
      {
         target[idx] = source.duplicate();
      }
      else
      {
         (target[idx]).assign(source);
      }
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   implements Scopeable
   {
      /** Stack of real dynamic-extent arrays, per each scope.*/
      private Deque<Set<BaseDataType[]>> dynamicArrays = new ArrayDeque<Set<BaseDataType[]>>();
      
      /** Set of all dynamic-extent arrays. */
      private Set<BaseDataType[]> allDynamicArrays = new HashSet<BaseDataType[]>();
      
      @Override
      public void scopeStart()
      {
         dynamicArrays.push(new HashSet<BaseDataType[]>());
      }

      @Override
      public void scopeFinished()
      {
         if (!dynamicArrays.isEmpty())
         {
            allDynamicArrays.removeAll(dynamicArrays.pop());
         }
      }
      
      /**
       * Deregister the given array as a dynamic-extent array.
       * 
       * @param    array
       *           The dynamic-extent array reference, must not be null.
       */
      public void deregister(BaseDataType[] array)
      {
         allDynamicArrays.remove(array);
         // go through all scopes and remove it
         for (Set<BaseDataType[]> arrays : dynamicArrays)
         {
            arrays.remove(array);
         }
      }
      
      /**
       * Register the given array as a dynamic-extent array.
       * 
       * @param    array
       *           The dynamic-extent array reference, must not be null.
       */
      public void register(BaseDataType[] array)
      {
         dynamicArrays.peek().add(array);
         allDynamicArrays.add(array);
      }
   }
}
